<?php

namespace Leenooks\SQRL;

use App\Http\Controllers\Controller;

use Illuminate\Http\Request;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Response;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Auth;
use Illuminate\Support\Facades\Log;
use Illuminate\View\View;

use Leenooks\SQRL\SQRL as SQRLAuth;

/**
 * Class SQRLController
 * @package Leenooks\SQRL
 * @todo CHECK THAT WE ARE RECEIVING BACK WHAT WE GIVE TO THE CLIENT
 * 	* QUERY: = sqrl url, eg: sqrl://domain/api/sqrl?nut=84cc3ef3b58b01dbe22931d1ceabdd6be2c27a481516755757c53b0162287bb8
 * 	* IDENT: = RESPONSE TO QUERY
 * @todo JOB TO DELETE OLD NONCES
 */

class SQRLController extends Controller
{
	/**
	 * Return our login page
	 *
	 * @param Request $request
	 * @return array|View
	 */
	public function auth(Request $request)
	{
		// Validate the nonce if it has been given.
		if ($request->get('nut')) {

		}

		// If this laravel, check if the user has been logged in
		if (app() instanceof \Illuminate\Foundation\Application) {
			// Laravel - check if the user is actually logged in
			if (Auth::check()) {
				return redirect()->intended(RouteServiceProvider::HOME);

			} else {
				return view('login',SQRLAuth::authNonce());
			}

		} else {
			// For JSON we just need the SQRL login
			if ($request->expectsJson())
				return Arr::get(SQRLAuth::authNonce(),'sqrl_url');

			return view('sqrl::login',SQRLAuth::authNonce())
				->with('LUMEN',TRUE);
		}
	}

	public function api(Request $request): Response
	{
		Log::info(sprintf('API-start [%s]',$request->get('nut')));

		$validatedData = $this->validate($request,[
			'client' => 'required|string',
			'server' => 'required|string',
			'ids' => 'required|string',
			'nut' => sprintf('required|string|exists:%s.nonces,nonce',config('sqrl.database')),
			'urs' => 'string',
			'pids' => 'string'
		]);

		Log::debug(sprintf('API-client [%s]',$validatedData['client']));
		Log::debug(sprintf('API-server [%s]',$validatedData['server']));
		Log::debug(sprintf('API-ids    [%s]',$validatedData['ids']));
		Log::debug(sprintf('API-urs    [%s]',$validatedData['urs'] ?? ''));
		Log::debug(sprintf('API-pids   [%s]',$validatedData['pids'] ?? ''));

		$decode_request = SQRL::decodeData($validatedData);
		$sqrl_nonce = SQRLAuth\Nonce::checkNonceValid($validatedData['nut']);
		$tif = in_array('noiptest',Arr::get($decode_request,'client.opt')) ? 0 : SQRL::tifcode('IP_MATCH');

		if (! $sqrl_nonce) {
			Log::error('API:Nonce not valid',['n'=>$validatedData['nut'],'tif'=>SQRL::tifcode('CLIENT_FAILURE')]);
			$response = SQRLAuth\Response::problem($validatedData['nut'],SQRL::tifcode('CLIENT_FAILURE'));

		} elseif (($sqrl_nonce->ip !== $request->ip()) && (! in_array('noiptest',Arr::get($decode_request,'client.opt')))) {
			Log::error('API::IP Doesnt Match',['n'=>$validatedData['nut'],'tif'=>SQRL::tifcode('COMMAND_FAILED')]);
			$response = SQRLAuth\Response::problem($sqrl_nonce->nonce,SQRL::tifcode('COMMAND_FAILED'));

		} elseif (SQRL::validateSignatures($validatedData,$decode_request) === FALSE) {
			Log::error('API::Signature Failed',['n'=>$validatedData['nut'],'tif'=>$tif|SQRL::tifcode('CLIENT_FAILURE')]);
			$response = SQRLAuth\Response::problem($sqrl_nonce->nonce,$tif|SQRL::tifcode('CLIENT_FAILURE'));

		} elseif (! Arr::get($decode_request,'client.cmd')) {
			Log::error('API::Invalid Request',['n'=>$validatedData['nut'],'tif'=>$tif|SQRL::tifcode('CLIENT_FAILURE')]);
			$response = SQRLAuth\Response::problem($sqrl_nonce->nonce,$tif|SQRL::tifcode("CLIENT_FAILURE"));

		} else {
			foreach (['ver','cmd'] as $y)
				Log::debug(sprintf('API-client-%s [%s]',str_pad($y,5,' '),Arr::get($decode_request,'client.'.$y)));
			Log::debug(sprintf('API-client-opt   [%s]',join('|',Arr::get($decode_request,'client.opt'))));
			Log::debug(sprintf('API-client-idk   [%s]',base64_encode(Arr::get($decode_request,'client.idk'))));
			Log::debug(sprintf('API-server       [%s]',serialize(Arr::get($decode_request,'server'))));

			Log::debug(sprintf('API-type   [%s]',$sqrl_nonce->type));

			switch ($sqrl_nonce->type) {
				case 'auth':
					$response = SQRLAuth\Auth::process(Arr::get($decode_request,'client'),Arr::get($decode_request,'server'),$sqrl_nonce,$tif);
					break;

				case 'question':
					$response = QuestionSQRLController::processRequestSQRL($decode_request["client"], $decode_request["server"], $sqrl_nonce, $validatedData["nut"], $tif);
					break;

				default:
					Log::error(__METHOD__.':Nonce Type Unsupported',['n'=>$validatedData['nut'],'tif'=>$tif|SQRL::tifcode('CLIENT_FAILURE')]);
					$response = SQRLAuth\Response::problem($sqrl_nonce->nonce,$tif|SQRL::tifcode('COMMAND_FAILED'));
			}
		}

		return SQRLAuth\Response::create($response);
	}

	/**
	 * Checks if nonce is valid, if nonce is valid the next page url
	 * or user response to a question is returned.
	 *
	 * @param Request $request
	 * @return JsonResponse
	 */
	public static function isReady(Request $request): JsonResponse
	{
		if ($request->get('nut')) {
			$o = SQRLAuth\Nonce::check($request->get('nut'),'orig_nonce');

			// If the nonce is old or doesnt exist.
			if (! $o) {
				Log::debug(sprintf('isReady: Invalid Nonce [%s]',$request->get('nut')));

				return response()->json([
					'isReady'=>FALSE,
					'msg'=>'Invalid Nonce, or Nonce expired'
				],200);
			}

			// Validate the IP matches - since the request would come from the same device client
			if ($o->ip !== $request->ip()) {
				Log::debug(sprintf('isReady: IP Mismatch [%s] != [%s]',$o->ip,$request->ip));

				return response()->json([
					'isReady'=>FALSE,
					'msg' => 'IP Mismatch',
				],200);
			}

			// Has the nonce be validated
			if ($o->verified != 1) {
				Log::debug(sprintf('isReady: Not Verified [%s]',$o->verified));

				return response()->json([
					'isReady'=>FALSE,
					'msg'=>'Not Ready'
				],404);
			}

			if ($o->pubkey && $o->pubkey->disabled) {
				Log::debug(sprintf('isReady: SQRL Disabled [%s]',$o->pubkey));

				return response()->json([
					'isReady'=>FALSE,
					'msg'=>'SQRL disabled for user'
				],200);
			}

			switch ($o->type) {
				case 'auth':
					Log::debug(sprintf('isReady: Authenticated [%s]',$o->pubkey));

					return response()->json([
						'isReady'=>TRUE,
						'msg'=>'SQRL authenticated',
						'nextPage'=>SQRL::base64_decode_url($o->url),
					],200);

				case 'question':
					break;
			}

		} else {
			Log::debug(sprintf('isReady: Not Nut?',$request->get('nut')));

			return response()->json([
				'isReady'=>FALSE,
				'msg'=>'Not Found!'
			],200);
		}
	}
}
